from useintest.executables.builders import CommandsBuilder
from useintest.executables.common import get_all_path_like_arguments_for_mounting
from useintest.executables.controllers import DefinedExecutablesControllerTypeBuilder
from useintest.executables.models import Executable

Samtools1_3_1ExecutablesController = DefinedExecutablesControllerTypeBuilder(
    "Samtools1_3_1ExecutablesController",
    {
        "samtools": Executable(
            CommandsBuilder(
                image="comics/samtools:1.3.1",
                executable="samtools",
                # There are so many ways in which Samtools accepts file paths that the creation of a parser for it would
                # be a massive task. Instead, we'll overzealously bind mount anything that looks like a file path.
                get_path_arguments_to_mount=get_all_path_like_arguments_for_mounting
            ),
            False
        )
    }
).build()

SamtoolsExecutablesController = Samtools1_3_1ExecutablesController

samtools_executable_controllers = {Samtools1_3_1ExecutablesController}



import atexit
import os
from copy import deepcopy

from docker.errors import NotFound
from temphelpers import TempManager
from typing import Dict, Optional, Type
from uuid import uuid4

from useintest.common import MOUNTABLE_TEMP_DIRECTORY, docker_client
from useintest.executables.builders import CommandsBuilder
from useintest.executables.common import CLI_ARGUMENTS, write_commands, pull_docker_image
from useintest.executables.models import Executable

_TAB_AS_SPACES = "    "


def _reduce_whitespace(string: str) -> str:
    """
    Reduces the whitespace in the given command.
    :param string: the command to reduce whitespace from
    :return: command with reduced whitespace
    """
    leading_tabs = int(min({(len(line) - len(line.lstrip(_TAB_AS_SPACES))) / len(_TAB_AS_SPACES)
                            for line in string.split("\n") if len(line.strip()) > 0}))
    stripped = []
    for line in string.split("\n"):
        stripped.append(line.replace(_TAB_AS_SPACES, "", leading_tabs))
    return "\n".join(stripped)


class ExecutablesController:
    """
    Controller for proxy executables that execute commands in a transparent Docker container.
    """
    def __init__(self, run_container_commands_builder: Optional[CommandsBuilder]=None):
        """
        Constructor.
        :param image_with_real_binaries: the name (docker-py's "tag") of the Docker image that the proxied binaries are
        executed within
        :param run_container_commands_builder: (optional) builder for commands used to start up persistent container in
        which commands should be run (can lead to much better performance because new container is not brought up each
        time)
        """
        self.run_container_command_builder = run_container_commands_builder

        if run_container_commands_builder is not None:
            if run_container_commands_builder.image is None:
                raise ValueError("Run container command builder must define the image the container is to use")
            pull_docker_image(run_container_commands_builder.image)
            self._cached_container_name = f"execution-container-{uuid4()}"
            self.run_container_command_builder.name = self._cached_container_name
            self.run_container_command_builder.detached = True

        atexit.register(self.tear_down)

    def tear_down(self):
        """
        Tears down the controller.
        """
        if self.run_container_command_builder is not None:
            try:
                container = docker_client.containers.get(self._cached_container_name)
                container.stop()
                container.remove(force=True)
            except NotFound:
                """ Not concerned if the container had not yet been created """

    def create_executable_commands(self, executable: Executable) -> str:
        """
        Creates executable commands for the given executable.
        :param executable: the executable to create comamnds for
        :return: the created commands
        """
        if executable.uses_running_container:
            if self.run_container_command_builder is None:
                raise ValueError("No command to run execution container defined.")
            executable.commands_builder.container = self._cached_container_name

            return _reduce_whitespace("""
                isRunning() {
                    [ $(docker ps -f name=%(uuid)s | wc -l | awk '{print $1}') -eq 2 ]
                }

                if ! isRunning;
                then
                    startIfNotRunning() {
                        if ! isRunning
                        then
                            %(container_setup)s > /dev/null
                        fi
                    }
                    lock=".%(uuid)s.lock"
                    if type flock > /dev/null 2>&1
                    then
                        # Linux
                        (
                            flock 10
                            startIfNotRunning
                            rm -f /tmp/${lock}
                        ) 10> /tmp/${lock}
                    elif type lockfile > /dev/null 2>&1
                    then
                        # Mac
                        lockfile /tmp/${lock}
                        startIfNotRunning
                        rm -f /tmp/${lock}
                    else
                        # No supported lock functionality - blindly try to start it and ignore any error
                        set +e
                        startIfNotRunning 2> /dev/null
                        set -e
                    fi
                fi

                %(to_execute)s
            """ % {
                "uuid": self._cached_container_name,
                "container_setup": self.run_container_command_builder.build().rstrip(),
                "to_execute": executable.commands_builder.build()
            })
        else:
            pull_docker_image(executable.commands_builder.image)
            return executable.commands_builder.build()

    def create_simple_executable_commands(self, containerised_executable: str, executable_arguments=CLI_ARGUMENTS) -> str:
        """
        TODO
        :param containerised_executable:
        :param executable_arguments:
        :return:
        """
        containerised_executable = containerised_executable.replace('"', '\\"')
        to_execute = CommandsBuilder(executable=containerised_executable, container=self._cached_container_name,
                                     executable_arguments=[executable_arguments])
        return self.create_executable_commands(Executable(to_execute, True))


class DefinedExecutablesController(ExecutablesController):
    """
    TODO
    """
    def __init__(self, run_container_commands_builder: Optional[CommandsBuilder]=None,
                 named_executables: Dict[str, Executable]=None):
        super().__init__(run_container_commands_builder)
        self._temp_manager = TempManager()
        self.named_executables = named_executables if named_executables is not None else dict()

    def tear_down(self):
        """
        TOOD
        :return:
        """
        super().tear_down()
        self._temp_manager.tear_down()

    def write_executables(self, location: str=None) -> str:
        """
        Writes the defined executables to the given location. If no location is given, they shall be written to a
        temporary directory.
        :param location: (optional) location to write the executables to
        :return: the directory containing the executables
        """
        if location is None:
            location = self._temp_manager.create_temp_directory(prefix="executables-", dir=MOUNTABLE_TEMP_DIRECTORY)

        for name, executable in self.named_executables.items():
            executable_location = os.path.join(location, name)
            commands = self.create_executable_commands(executable)
            write_commands(executable_location, commands)

        return location


class DefinedExecutablesControllerTypeBuilder:
    """
    TODO
    """
    def __init__(self, type_name: str, named_executables: Dict[str, Executable]):
        """
        TODO
        :param type_name:
        :param named_executables:
        """
        self.type_name = type_name
        self.named_executables = named_executables

    def build(self) -> Type[DefinedExecutablesController]:
        """
        TODO
        :return:
        """
        named_executables = deepcopy(self.named_executables)

        def init(controller: DefinedExecutablesController, *args, **kwargs):
            super(type(controller), controller).__init__(*args, named_executables=named_executables, **kwargs)

        return type(
            self.type_name,
            (DefinedExecutablesController, ),
            {"__init__": init}
        )



from bidict import bidict
from docker.errors import NotFound
from docker.models.containers import Container
from typing import Set, Optional

from useintest.common import UseInTestModel, docker_client
from useintest.services.exceptions import UnexpectedNumberOfPortsError


class Service(UseInTestModel):
    """
    Model of a service.
    """
    def __init__(self):
        """
        Constructor.
        """
        super().__init__()
        # FIXME: Assumption about where the Docker machine is accessible (i.e. it could be on a VM)
        self.host = "localhost"
        self.ports = bidict()

    @property
    def port(self) -> int:
        """
        Gets the port on the localhost. If there is more than one port exposed, an
        `UnexpectedNumberOfPortsException` will be raised.
        :return: the exposed port
        """
        if len(self.ports) != 1:
            raise UnexpectedNumberOfPortsError("%d ports are exposed (cannot use `port`)" % len(self.ports))
        return list(self.ports.values())[0]

    def get_external_port_mapping_to(self, port: int) -> int:
        """
        Gets the port on localhost to which the given port in the container maps to.
        :param port: the port inside the container
        :return: the port outside that maps to that in the container
        """
        return self.ports.inv[port]


class DockerisedService(Service):
    """
    Model of a service running in a Docker container.
    """
    @property
    def container(self) -> Optional[Container]:
        if self.container_id is None:
            return None
        try:
            return docker_client.containers.get(self.container_id)
        except NotFound:
            return None

    @container.setter
    def container(self, container: Container):
        self.container_id = container.id

    @property
    def container_id(self) -> Optional[str]:
        return self._container_id

    @container_id.setter
    def container_id(self, container_id: str):
        self._container_id = container_id

    def __init__(self):
        super().__init__()
        self.name = None
        self._container_id: str = None
        self.controller = None

    # TODO: Not sure of the best way to specify the type as it could be that of a subclass...
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.controller.stop_service(self)


class User(UseInTestModel):
    """
    A user with an associated password.
    """
    def __init__(self, username: str, password: str=None):
        self.username = username
        self.password = password


class ServiceWithUsers(Service):
    """
    A service with users.
    """
    def __init__(self):
        super().__init__()
        self.users: Set[User] = set()
        self._root_user: Optional[User] = None

    @property
    def root_user(self) -> Optional[User]:
        """
        Gets a user of the service that has privileged access. 
        :return: a user with privilege access
        """
        assert self._root_user in self.users
        return self._root_user

    @root_user.setter
    def root_user(self, user: Optional[User]):
        """
        Sets the user of the service that has privileged access.
        :param user: the user with privilege access
        """
        if user is not None and not user in self.users:
            self.users.add(user)
        self._root_user = user


class DockerisedServiceWithUsers(DockerisedService, ServiceWithUsers):
    """
    Service running on Docker with users.
    """



import unittest
from abc import ABCMeta
from typing import Dict

from testhelpers import TypeUsedInTest, get_classes_to_test, create_tests

from useintest.modules.irods.helpers import IrodsSetupHelper, AccessLevel
from useintest.modules.irods.models import IrodsUser
from useintest.modules.irods.services import IrodsServiceController, irods_service_controllers
from useintest.modules.irods.setup_irods import setup_irods
from useintest.services.models import DockerisedServiceWithUsers
from useintest.tests.common import extract_version_number
from useintest.tests.services.common import TestServiceControllerSubclass

_METADATA = {
    "attribute_1": ["value_1", "value_2"],
    "attribute_2": ["value_3", "value_4"],
    "attribute_3": "value_5"
}
_DATA_OBJECT_NAME = "data-object-name"


class _TestIrodsSetupHelper(
    TestServiceControllerSubclass[TypeUsedInTest, DockerisedServiceWithUsers], metaclass=ABCMeta):
    """
    Tests for `IrodsSetupHelper`.
    """
    def setUp(self):
        super().setUp()
        self.icommands_location, self.service, self.icommands_controller, self.icat_controller = setup_irods(
            self.get_type_to_test())
        self.setup_helper = IrodsSetupHelper(self.icommands_location)

    def tearDown(self):
        self.icat_controller.stop_service(self.service)
        self.icommands_controller.tear_down()

    def test_run_icommand(self):
        ils = self.setup_helper.run_icommand(["ils"])
        self.assertTrue(ils.startswith("/"))
        self.assertTrue(ils.endswith(":"))

    def test_create_data_object_with_path_opposed_to_name(self):
        self.assertRaises(ValueError, self.setup_helper.create_data_object, "/test")

    def test_create_data_object(self):
        contents = "Test contents"
        path = self.setup_helper.create_data_object(_DATA_OBJECT_NAME, contents=contents)
        self.setup_helper.run_icommand(["icd", path.rsplit('/', 1)[-1]])
        self.assertIn(_DATA_OBJECT_NAME, self.setup_helper.run_icommand(["ils"]))

    def test_get_data_object(self):
        contents = "Test contents"
        path = self.setup_helper.create_data_object(_DATA_OBJECT_NAME, contents=contents)
        self.assertEqual(contents, self.setup_helper.read_data_object(path))

    def test_replicate_data_object(self):
        data_object_location = self.setup_helper.create_data_object(_DATA_OBJECT_NAME)
        resource = self.setup_helper.create_replica_storage()
        self.setup_helper.replicate_data_object(data_object_location, resource)

        collection_listing = self.setup_helper.run_icommand(["ils", "-l"])
        self.assertIn("1 %s" % resource.name[0:20], collection_listing)

    def test_create_collection(self):
        collection_name = "collection"
        path = self.setup_helper.create_collection(collection_name)

        self.setup_helper.run_icommand(["icd", path.rsplit('/', 1)[-1]])
        self.assertIn(collection_name, self.setup_helper.run_icommand(["ils"]))

    def test_create_collection_with_collection_path_opposed_to_collection_name(self):
        self.assertRaises(ValueError, self.setup_helper.create_collection, "/test")

    def test_add_metadata_to_data_object(self):
        path = self.setup_helper.create_data_object(_DATA_OBJECT_NAME)

        self.setup_helper.add_metadata_to(path, _METADATA)

        retrieved_metadata = self.setup_helper.run_icommand(["imeta", "ls", "-d", path])
        self._assert_metadata_in_retrieved(_METADATA, retrieved_metadata)

    def test_add_metadata_to_collection(self):
        path = self.setup_helper.create_collection("collection")

        self.setup_helper.add_metadata_to(path, _METADATA)

        retrieved_metadata = self.setup_helper.run_icommand(["imeta", "ls", "-c", path])
        self._assert_metadata_in_retrieved(_METADATA, retrieved_metadata)

    def test_update_checksums(self):
        path = self.setup_helper.create_data_object(_DATA_OBJECT_NAME, "abc")
        resource = self.setup_helper.create_replica_storage()
        self.setup_helper.replicate_data_object(_DATA_OBJECT_NAME, resource)

        expected_checksum = "900150983cd24fb0d6963f7d28e17f72" if self.service.version.major == 3 \
            else "sha2:ungWv48Bz+pBQUDeXa4iI7ADYaOWF3qctBD/YfIAFa0="

        # Asserting that checksum is not stored before now
        assert expected_checksum not in self.setup_helper.run_icommand(["ils", "-L", path])
        self.setup_helper.update_checksums(path)

        ils = self.setup_helper.run_icommand(["ils", "-L", path])
        self.assertEqual(2, ils.count(expected_checksum))

    def test_get_checksum(self):
        path = self.setup_helper.create_data_object(_DATA_OBJECT_NAME, "abc")
        expected_checksum = "900150983cd24fb0d6963f7d28e17f72" if self.service.version.major == 3 \
            else "sha2:ungWv48Bz+pBQUDeXa4iI7ADYaOWF3qctBD/YfIAFa0="
        self.assertEqual(expected_checksum, self.setup_helper.get_checksum(path))

    def test_create_replica_storage(self):
        resource = self.setup_helper.create_replica_storage()
        resource_info = self.setup_helper.run_icommand(["iadmin", "lr", resource.name])
        self.assertIn("resc_name: %s" % resource.name, resource_info)
        self.assertIn("resc_def_path: %s" % resource.location, resource_info)

    def test_create_user_with_existing_username(self):
        existing_user = self.service.users[0]
        self.assertRaises(ValueError, self.setup_helper.create_user, existing_user.username, existing_user.zone)

    def test_create_user(self):
        expected_user = IrodsUser("user_1", self.service.users[0].zone)
        user = self.setup_helper.create_user(expected_user.username, expected_user.zone)
        self.assertEqual(expected_user, user)
        user_list = self.setup_helper.run_icommand(["iadmin", "lu"])
        self.assertIn("%s#%s" % (expected_user.username, expected_user.zone), user_list)

    def test_set_access(self):
        path = self.setup_helper.create_data_object(_DATA_OBJECT_NAME)
        zone = self.service.users[0].zone
        user_1 = self.setup_helper.create_user("user_1", zone)
        user_2 = self.setup_helper.create_user("user_2", zone)

        self.setup_helper.set_access(user_1.username, AccessLevel.READ, path)
        self.setup_helper.set_access(user_2.username, AccessLevel.WRITE, path)

        access_info = self.setup_helper.run_icommand(["ils", "-A", path])
        self.assertIn("%s#%s:read object" % (user_1.username, zone), access_info)
        self.assertIn("%s#%s:modify object" % (user_2.username, zone), access_info)

    def test_get_icat_version(self):
        self.assertEqual(self.service.version, self.setup_helper.get_icat_version())

    def _assert_metadata_in_retrieved(self, metadata: Dict, retrieved_metadata: str):
        """
        Assert that the given metadata is in the metadata information retrieved via an `imeta` command.
        :param metadata: the metadata to expect
        :param retrieved_metadata: string representation of metadata, retrieved via an `imeta` command
        """
        for attribute in metadata:
            attribute_values = metadata[attribute]
            if not isinstance(attribute_values, list):
                attribute_values = [attribute_values]

            for value in attribute_values:
                self.assertIn("attribute: %s\nvalue: %s" % (attribute, value), retrieved_metadata)


globals().update(create_tests(_TestIrodsSetupHelper, get_classes_to_test(irods_service_controllers, IrodsServiceController),
                              lambda superclass, test_type: "TestIrodsSetupHelperWithIrods%s"
                                                            % extract_version_number(test_type.__name__).replace(".", "_")))

# Fix for stupidity of test runners
del _TestIrodsSetupHelper, TestServiceControllerSubclass, create_tests, get_classes_to_test

if __name__ == "__main__":
    unittest.main()



import base64
import os
import sys
from typing import Set

from dill import dill

sys.path.append(os.path.join(os.path.dirname(os.path.realpath(__file__)), "../../"))

if __name__ == "__main__":
    if len(sys.argv) <= 2:
        print("")
        exit(0)

    serialised_parser = sys.argv[1]
    arguments = sys.argv[2:]

    parser = dill.loads(base64.b64decode(serialised_parser))
    mounts = parser(arguments)

    absolute_paths = False
    processed_mounts = set()    # type: Set[str]
    for path in mounts:
        if not os.path.isabs(path):
            absolute_paths = True
        if not os.path.exists(path) or os.path.isfile(path):
            path = os.path.dirname(path)
        path = os.path.abspath(path)
        if os.path.exists(path):
            # Helping wildcard mounts to work by checking for existence before adding to the mount set
            processed_mounts.add(path)

    if absolute_paths:
        # TODO: The correct thing to do here is modify all the arguments so they use absolute paths instead of relative
        # ones. However, this is too much of a parsing nightmare for me to be motivated to do it at the moment. This
        # alternative solution will break Docker images where the entrypoint uses relative paths (as the work directory
        # would have been changed).
        print("-w %s " % os.path.abspath(""), end="")

    print(" ".join(["-v {mount}:{mount}".format(mount=mount) for mount in processed_mounts]), end="")



from abc import ABCMeta
from typing import Set, Generic, TypeVar
from unittest import TestCase

from testhelpers import TypeUsedInTest, TestUsingType

from useintest.common import docker_client
from useintest.services.models import DockerisedService, Service

ServiceType = TypeVar("ServiceType", bound=Service)


# TODO: These need sorting out - why is there 2 classes here?
class TestServiceControllerSubclass(Generic[TypeUsedInTest, ServiceType], TestUsingType[TypeUsedInTest], TestCase,
                                    metaclass=ABCMeta):
    """
    TODO
    """
    def setUp(self):
        super().setUp()
        self._started: Set[ServiceType] = set()
        self.service_controller = type(self).get_type_to_test()()

    def tearDown(self):
        for service in self._started:
            self.service_controller.stop_service(service)

    def _start_service(self) -> ServiceType:
        service = self.service_controller.start_service()
        self._started.add(service)
        return service


class TestDockerisedServiceControllerSubclass(Generic[TypeUsedInTest, ServiceType],
                                              TestServiceControllerSubclass[TypeUsedInTest, ServiceType],
                                              metaclass=ABCMeta):
    """
    Superclass for `DockerisedServiceController` tests.
    """
    def test_stop(self):
        service = self._start_service()
        assert len(docker_client.containers.list(filters=dict(name=service.name))) == 1
        self.service_controller.stop_service(service)
        self.assertEqual(0, len(docker_client.containers.list(filters=dict(name=service.name))))

    def test_stop_when_not_started(self):
        service = DockerisedService()
        self.service_controller.stop_service(service)



from useintest.modules.gitlab.gitlab import ROOT_USERNAME, ROOT_PASSWORD, GitLabServiceController, \
    gitlab_service_controllers, GitLab8_10_4_ce_0ServiceController, GitLab8_13_11_ce_0ServiceController, \
    GitLab8_16_6_ce_0ServiceController
from useintest.modules.gitlab.helpers import SshKey



import os

from useintest.common import MissingDependencyError
from useintest.services.builders import DockerisedServiceControllerTypeBuilder
from useintest.services.models import DockerisedService

DEFAULT_HTTP_PORT = 8500

_repository = "consul"
_ports = [8300, 8301, 8302, DEFAULT_HTTP_PORT, 8600]
_start_detector = lambda log_line: "Node info in sync" in log_line or "Synced node info" in log_line


class ConsulDockerisedService(DockerisedService):
    """
    Consul service.
    """
    CONSUL_ADDRESS_ENVIRONMENT_VARIABLE = "CONSUL_HTTP_ADDR"
    CONSUL_TOKEN_ENVIRONMENT_VARIABLE = "CONSUL_HTTP_TOKEN"
    CONSUL_SCHEME_ENVIRONMENT_VARIABLE = "CONSUL_SCHEME"
    CONSUL_DATACENTRE_ENVIRONMENT_VARIABLE = "CONSUL_DC"
    CONSUL_VERIFY_ENVIRONMENT_VARIABLE = "CONSUL_HTTP_SSL_VERIFY"
    CONSUL_CERTIFICATE_ENVIRONMENT_VARIABLE = "CONSUL_CLIENT_CERT"

    @staticmethod
    def _clear_environment():
        """
        Clears the environment variables related to Consul.
        """
        os.environ.pop(ConsulDockerisedService.CONSUL_ADDRESS_ENVIRONMENT_VARIABLE, None)
        os.environ.pop(ConsulDockerisedService.CONSUL_TOKEN_ENVIRONMENT_VARIABLE, None)
        os.environ.pop(ConsulDockerisedService.CONSUL_SCHEME_ENVIRONMENT_VARIABLE, None)
        os.environ.pop(ConsulDockerisedService.CONSUL_DATACENTRE_ENVIRONMENT_VARIABLE, None)
        os.environ.pop(ConsulDockerisedService.CONSUL_VERIFY_ENVIRONMENT_VARIABLE, None)
        os.environ.pop(ConsulDockerisedService.CONSUL_CERTIFICATE_ENVIRONMENT_VARIABLE, None)

    def create_consul_client(self):
        """
        Gets a client for the Consul service.

        Requires consul (not installed with useintest)
        :return: the consul client
        """
        try:
            from consul import Consul
        except ImportError as e:
            raise MissingDependencyError("python-consul") from e
        ConsulDockerisedService._clear_environment()
        return Consul(self.host, self.ports[DEFAULT_HTTP_PORT])

    def setup_environment(self):
        """
        Sets Consul related environment variables.
        """
        ConsulDockerisedService._clear_environment()
        os.environ[ConsulDockerisedService.CONSUL_ADDRESS_ENVIRONMENT_VARIABLE] = \
            f"{self.host}:{self.ports[DEFAULT_HTTP_PORT]}"


common_setup = {
    "repository": _repository,
    "start_detector": _start_detector,
    "ports": _ports,
    "service_model": ConsulDockerisedService
}

Consul1_0_0ServiceController = DockerisedServiceControllerTypeBuilder(
    name="Consul1_0_0ServiceController",
    tag="1.0.0",
    **common_setup).build()

Consul0_8_4ServiceController = DockerisedServiceControllerTypeBuilder(
    name="Consul0_8_4ServiceController",
    tag="0.8.4",
    **common_setup).build()


ConsulServiceController = Consul1_0_0ServiceController

consul_service_controllers = {Consul1_0_0ServiceController, Consul0_8_4ServiceController}



import os
import subprocess
from typing import Sequence, Type

from useintest.executables.builders import CommandsBuilder, MountedArgumentParserBuilder
from useintest.executables.controllers import DefinedExecutablesController
from useintest.executables.models import Executable
from useintest.modules.irods.models import Version


class IrodsBaseExecutablesController(DefinedExecutablesController):
    """
    Executables (icomands) for use against an iRODS server (iCAT).
    """
    _ICOMMAND_EXECUTABLES = {"ibun", "icd", "ichksum", "ichmod", "icp", "idbug", "ienv", "ierror", "iexecmd", "iexit",
                             "ifsck", "iget", "igetwild", "ihelp", "iinit", "ilocate", "ils", "ilsresc", "imcoll",
                             "imiscsvrinfo", "imkdir", "imv", "ipasswd", "iphybun", "iphymv", "ips", "iput", "ipwd",
                             "iqdel", "iqmod", "iqstat", "iquest", "iquota", "ireg", "irepl", "irm", "irmtrash",
                             "irsync", "irule", "iscan", "isysmeta", "itrim", "iuserinfo", "ixmsg", "izonereport",
                             "imeta", "iadmin"}
    _GET_POSITIONAL_ARGUMENTS_TO_MOUNT = MountedArgumentParserBuilder(
        positional_arguments=MountedArgumentParserBuilder.ALL_POSITIONAL_ARGUMENTS).build()
    _DEFAULT_SETTINGS_DIRECTORIES = {"/root/.irods", "/home/root/.irods"}

    # TODO: Could add option to connect to iRODS server not running in Docker (i.e. via port opposed to link)
    def __init__(self, irods_container_name: str, image_with_compatible_icommands: str, settings_directory_on_host: str,
                 settings_directories_in_container: Sequence[str]=_DEFAULT_SETTINGS_DIRECTORIES):
        """
        Constructor.
        :param irods_container_name: the name of the container running the iRODS server
        :param image_with_compatible_icommands: image containing icommands that are compatible with the iRODS server
        :param settings_directory_on_host: directory on the Docker host machine that are used to access iRODS
        :param settings_directories_in_container: the directories on the container running the Docker image that need to
        contain the settings
        """
        self._image_with_compatible_icommands = image_with_compatible_icommands
        self._run_container_commands_builder = CommandsBuilder(
            "sleep", executable_arguments=["infinity"], image=image_with_compatible_icommands,
            other_docker="--link %s" % irods_container_name,
            mounts={settings_directory_on_host: set(settings_directories_in_container)})
        super().__init__(run_container_commands_builder=self._run_container_commands_builder)
        self._register_named_executables()

    def authenticate(self, executables_directory: str, password: str):
        """
        Authenticate "client" with the iRODS server using `iinit` in the given executables directory and the given
        password.
        :param executables_directory: the directory containing the iRODS executables
        :param password: the password used to authenticate based on the settings provided to the constructor
        """
        process = subprocess.Popen([os.path.join(executables_directory, "iinit"), password], stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        out, error = process.communicate()
        if len(error) > 0:
            raise Exception("Error authenticating to iCAT server with password \"%s\": %s" % (password, error))

    def write_executables_and_authenticate(self, password: str, location: str=None) -> str:
        """
        Both writes the executables to the given location and then authenticates to use the server with the given
        password.
        :param password: password used by the `authenticate` method
        :param location: location used by the `write_executables` method
        :return: the location of the written executables
        """
        location = self.write_executables(location)
        self.authenticate(location, password)
        return location

    def _register_named_executables(self):
        """
        Registers the executables that can be written by this controller.
        """
        for icommand in IrodsBaseExecutablesController._ICOMMAND_EXECUTABLES - {"iget", "iput"}:
            self.named_executables[icommand] = Executable(CommandsBuilder(icommand), True)

        def create_executable_template(command: str):
            commands_builder = CommandsBuilder(
                command, image=self._image_with_compatible_icommands,
                get_path_arguments_to_mount=IrodsBaseExecutablesController._GET_POSITIONAL_ARGUMENTS_TO_MOUNT,
                mounts=self._run_container_commands_builder.mounts,
                other_docker=self._run_container_commands_builder.other_docker)
            return Executable(commands_builder, False)

        # Note: if `-` is the second positional argument with `iget`, `-` is suspected as a file, relative to the
        # current directory. This leads to an unnecessary mount and the use of `-w` to change the working directory.
        self.named_executables["iget"] = create_executable_template("iget")
        self.named_executables["iput"] = create_executable_template("iput")


def _build_irods_executables_controller(image_with_compatible_icommands: str, irods_version: Version) \
        -> Type[IrodsBaseExecutablesController]:
    """
    TODO
    :param image_with_compatible_icommands:
    :param irods_version:
    :return:
    """
    def init(self, *args, **kwargs):
        args = list(args)
        args.insert(1, image_with_compatible_icommands)
        args = tuple(args)
        super(type(self), self).__init__(*args, **kwargs)

    return type(
        "Irods%sExecutablesController" % str(irods_version).replace(".", "_"),
        (IrodsBaseExecutablesController,),
        {"__init__": init}
    )


Irods3_3_1ExecutablesController = _build_irods_executables_controller("mercury/icat:3.3.1", Version("3.3.1"))
Irods4_1_8ExecutablesController = _build_irods_executables_controller("mercury/icat:4.1.8", Version("4.1.8"))
Irods4_1_9ExecutablesController = _build_irods_executables_controller("mercury/icat:4.1.9", Version("4.1.9"))
Irods4_1_10ExecutablesController = _build_irods_executables_controller("mercury/icat:4.1.10", Version("4.1.10"))
IrodsExecutablesController = Irods4_1_10ExecutablesController

irods_executables_controllers_and_versions = {
    Version("3.3.1"): Irods3_3_1ExecutablesController,
    Version("4.1.8"): Irods4_1_8ExecutablesController,
    Version("4.1.9"): Irods4_1_9ExecutablesController,
    Version("4.1.10"): Irods4_1_10ExecutablesController,
}
irods_executables_controllers = irods_executables_controllers_and_versions.values()



import subprocess
from typing import List, Tuple, Optional

from useintest.tests.common import MAX_RUN_TIME_IN_SECONDS
from useintest.executables.builders import CommandsBuilder

UBUNTU_IMAGE_TO_TEST_WITH = "ubuntu:16.04"


def get_builder_for_commands_to_run_persistent_ubuntu() -> CommandsBuilder:
    """
    Gets commands builder to run a persistent Ubuntu container.
    :return: the commands builder
    """
    return CommandsBuilder("sleep", image=UBUNTU_IMAGE_TO_TEST_WITH, executable_arguments=["infinity"])


def run(arguments: List=None, raise_if_stderr: bool=True, pipe_in: str=None, decode_output_to: Optional[str]="utf-8") \
        -> Tuple[str, str]:
    """
    Calls out to run the given arguments on the system.
    :param arguments: the arguments to execute, where the first item is the executable and subsequent items are the
    executable's arguments
    :param raise_if_stderr: whether the test should raise an exception if anything is written to standard error
    :param pipe_in: any string that should be piped into the process as input
    :param decode_output_to: the output character encoding
    :return: tuple where the first element is what was written to standard out and the second is what was written to
    standard error
    """
    if pipe_in is not None and isinstance(pipe_in, str):
        pipe_in = str.encode(pipe_in)

    process = subprocess.Popen(arguments, stdout=subprocess.PIPE, stdin=subprocess.PIPE, stderr=subprocess.PIPE)
    out, error = process.communicate(input=pipe_in, timeout=MAX_RUN_TIME_IN_SECONDS)
    if decode_output_to is not None:
        out = out.decode(decode_output_to).rstrip("\n")
        error = error.decode(decode_output_to)

    if raise_if_stderr and len(error) > 0:
        raise ValueError("Unexpected output on standard error:\n%s" % error)

    return out, error



# FIXME: This is not cross platform...
from abc import ABCMeta

import docker

MOUNTABLE_TEMP_DIRECTORY = "/tmp"

docker_client = docker.from_env()


class UseInTestError(Exception):
    """
    Base class for all custom exceptions defined in this package.
    """


class MissingDependencyError(UseInTestError):
    """
    Raised when an optional package is not installed.
    """
    def __init__(self, package_name: str):
        super().__init__(f"Optional Python dependency `{package_name}` must be installed separately to use this "
                         f"functionality")


class UseInTestModel(metaclass=ABCMeta):
    """
    Base which all models in this package should extend.
    """





import atexit
import logging
import os
import platform
import subprocess
from enum import Enum, unique
from tempfile import mkdtemp

import re
import shutil
from typing import List, Union, Dict, Set
from uuid import uuid4

from useintest.modules.irods.models import IrodsResource, IrodsUser, Version


@unique
class AccessLevel(Enum):
    """
    Entity access levels available in iRODS.
    """
    NONE = "null"
    READ = "read"
    WRITE = "write"
    OWN = "own"


class IrodsSetupHelper:
    """
    Helper for setting up tests using iRODS.
    """
    def __init__(self, icommands_location: str):
        """
        Constructor.
        :param icommands_location: the location of the icommands that can be used to communicate with the iRODS server
        """
        self.icommands_location = icommands_location

    def create_data_object(self, name: str, contents: str="") -> str:
        """
        Creates a test data object on iRODS with the given name and contents.
        :param name: the nane of the file to create
        :param contents: the contents of the file to create
        :return: the path to the created file
        """
        if "/" in name:
            raise ValueError("Data object name cannot include '/'")

        # TODO: Consider using `TempManager` from `hgicommon` instead
        def remove_temp_folder(location: str):
            if os.path.exists(location):
                try:
                    shutil.rmtree(location)
                except OSError:
                    pass

        temp_root = None
        if platform.system() == "Darwin":
            # The temp directory on a Mac machine is somewhere that cannot be bind mounted by Docker using default
            # settings. /tmp can however
            temp_root = "/tmp"

        temp_directory_path = mkdtemp(dir=temp_root)
        atexit.register(remove_temp_folder, temp_directory_path)

        temp_file_path = os.path.join(temp_directory_path, name)
        os.chmod(temp_directory_path, 0o770)

        with open(temp_file_path, "w+") as temp_file:
            temp_file.write(contents)
        os.chmod(temp_file_path, 0o770)

        self.run_icommand(["iput", temp_file_path])
        remove_temp_folder(temp_directory_path)
        atexit.unregister(remove_temp_folder)

        return "%s/%s" % (self.run_icommand(["ipwd"]), name)

    def read_data_object(self, path: str) -> str:
        """
        Reads the contents of the entity with the given path.
        :param path: the path to the entity in iRODS
        :return: the entity's contents
        """
        return self.run_icommand(["iget", path, "-"])

    def replicate_data_object(self, path: str, replicate_to: Union[str, IrodsResource]):
        """
        Replicates the data object in the given path to the given resource.
        :param path: the path of the data object that is to be replicated
        :param replicate_to: the resource or name of the resource to which the data object should be replicated to
        """
        if isinstance(replicate_to, IrodsResource):
            replicate_to = replicate_to.name
        self.run_icommand(["irepl", "-R", replicate_to, path])

    def create_collection(self, name: str) -> str:
        """
        Creates a test collection on iRODS with the given name and contents.
        :param name: the name of the collection to create
        :return: the path to the created collection
        """
        if "/" in name:
            raise ValueError("Collection name cannot include '/'")

        self.run_icommand(["imkdir", name])

        return "%s/%s" % (self.run_icommand(["ipwd"]), name)

    def add_metadata_to(self, path: str, metadata: Dict):
        """
        Adds the given metadata to the entity at the given path in iRODS.
        :param path: the path to add metadata to (could correspond to a collection or data object)
        :param metadata: the metadata to add
        """
        if len(metadata) > 0:
            type_flag = "-c" if self.is_collection(path) else "-d"

            for key, values in metadata.items():
                if not isinstance(values, List) and not isinstance(values, Set):
                    values = [values]
                assert type(values) != str
                for value in values:
                    self.run_icommand(["imeta", "add", type_flag, path, key, str(value)])

    def is_collection(self, path: str) -> bool:
        """
        Checks whether the given path in iRODS is a collection.
        :param path: the path to check
        :return: whether there is a collection at the given path
        """
        listing = self.run_icommand(["ils", path])
        return ":" in listing

    def update_checksums(self, path: str):
        """
        Forces iRODS to update the checksums of all replicas of the data object with the path given/all data objects
        in the collection given (recursive).
        :param path: the path to the data object/collection
        """
        self.run_icommand(["ichksum", "-f", "-a", "-r", path])

    def get_checksum(self, path: str) -> str:
        """
        Gets the checksum of the most recently updated replica of a data object on iRODS.

        If not stored in iRODS, the checksum will be calculated and stored as an unavoidable side-effect.
        :param path: the path to the data object
        :return: the checksum of the data object
        """
        checksum_out = self.run_icommand(["ichksum", path])
        return checksum_out.split('\n')[0].rsplit(' ', 1)[-1]

    def create_replica_storage(self) -> IrodsResource:
        """
        Creates replica storage resource.
        :return: resource on which replicas can be stored
        """
        name = str(uuid4())
        location = "/tmp/%s" % name
        host = "localhost"
        if self.get_icat_version().major == 3:
            self.run_icommand(
                ["iadmin", "mkresc", "%s" % name, "unix file system", "cache", "%s" % host, "%s" % location])
        else:
            self.run_icommand(["iadmin", "mkresc", "%s" % name, "unixfilesystem", "%s:%s" % (host, location)])
        return IrodsResource(name, host, location)

    def create_user(self, username: str, zone: str) -> IrodsUser:
        """
        Creates a user with a given username in the given zone.
        :param username: the username the user should have
        :param zone: the zone the user should be in
        :return: the created user
        """
        user = IrodsUser(username, zone, None)
        try:
            self.run_icommand(["iadmin", "mkuser", "%s#%s" % (username, zone), "rodsuser"])
        except RuntimeError as e:
            if "CATALOG_ALREADY_HAS_ITEM_BY_THAT_NAME" in e.args[0]:
                raise ValueError("A user already exists with the given username")
        return user

    def set_access(self, user_or_group: str, level: AccessLevel, path: str):
        """
        Sets the given access level for a user or group on the entity at the given path in iRODS.
        :param user_or_group: the user or group which the access level is been set for
        :param level: the access level
        :param path: the path of the entity
        """
        self.run_icommand(["ichmod", level.value, user_or_group, path])

    def get_icat_version(self) -> Version:
        """
        Gets the version of iCAT server being used.
        :return: the version of iCAT server
        """
        ienv_out = self.run_icommand(["ienv"])
        version_as_string = re.search("rods(.*),", ienv_out).group(1)
        return Version(version_as_string)

    def run_icommand(self, arguments: Union[str, List[str]], deprecated_arguments: List[str]=None) -> str:
        """
        Executes the given icommand binary with any arguments, returning the stdout as a string and raising an
        exception if stderr is written to.
        :param arguments: the binary to execute (must be icommand binary with no path, e.g. ["ils", args]) and arguments
        :param deprecated_arguments: (deprecated - pass after binary in list given in first argument) command arguments
        :return: the output written to stdout by the icommand that was executed
        """
        if isinstance(arguments, str):
            logging.warning("Use of a string denoting the icommand binary with an optional list of arguments is "
                            "depreciated - combine both in a single list, given as the first argument")
            arguments = [arguments]
            if deprecated_arguments is not None:
                arguments += deprecated_arguments

        binary_path = os.path.join(self.icommands_location, arguments[0])
        arguments[0] = binary_path

        process = subprocess.Popen(arguments, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, error = process.communicate()
        logging.debug("icommand output: %s" % out)

        if len(error) != 0:
            raise RuntimeError("%s:\nError: %s\nOutput: %s" % (arguments, error, out))

        return out.decode("utf-8").rstrip()



import re

from useintest.common import MOUNTABLE_TEMP_DIRECTORY

MOUNTABLE_TEMP_CREATION_KWARGS = {"dir": MOUNTABLE_TEMP_DIRECTORY}

MAX_RUN_TIME_IN_SECONDS = 120

_EXTRACT_VERSION_PATTERN = re.compile("[0-9]+(_[0-9]+)*")


def extract_version_number(string: str) -> str:
    """
    Extracts a version from a string in the form: `.*[0-9]+(_[0-9]+)*.*`, e.g. Irods4_1_9CompatibleController.

    If the string contains multiple version numbers, the first (from left) is extracted.

    Will raise a `ValueError` if there is no version number in the given string.
    :param string: the string containing the version number
    :return: the extracted version
    """
    matched = _EXTRACT_VERSION_PATTERN.search(string)
    if matched is None:
        raise ValueError("No version number in string")
    return matched.group().replace("_", ".")



import unittest
from abc import ABCMeta

from couchdb import Server

from useintest.modules.couchdb import couchdb_service_controllers, CouchDBServiceController
from useintest.services.models import DockerisedServiceWithUsers
from testhelpers import TypeUsedInTest, create_tests, get_classes_to_test
from useintest.tests.services.common import TestDockerisedServiceControllerSubclass


class _TestCouchDBDockerisedServiceController(
    TestDockerisedServiceControllerSubclass[TypeUsedInTest, DockerisedServiceWithUsers], metaclass=ABCMeta):
    """
    Tests for CouchDB service controllers.
    """
    def test_start(self):
        service = self._start_service()
        couch = Server("http://%s:%d" % (service.host, service.port))
        database = couch.create("test-database")
        posted = {"this": "value"}
        identifier, revision = database.save(posted)
        self.assertEqual(posted, database[identifier])


# Setup tests
globals().update(create_tests(_TestCouchDBDockerisedServiceController, get_classes_to_test(couchdb_service_controllers, CouchDBServiceController)))


# Fix for stupidity of test runners
del _TestCouchDBDockerisedServiceController, TestDockerisedServiceControllerSubclass, create_tests, get_classes_to_test


if __name__ == "__main__":
    unittest.main()



import unittest
from abc import ABCMeta

from testhelpers import TestUsingType, create_tests, get_classes_to_test, TypeUsedInTest

from useintest.modules.irods.executables import IrodsBaseExecutablesController
from useintest.modules.irods.helpers import IrodsSetupHelper
from useintest.modules.irods.services import irods_service_controllers, IrodsServiceController
from useintest.modules.irods.setup_irods import setup_irods
from useintest.tests.common import extract_version_number


class _TestSetupIrods(TestUsingType[TypeUsedInTest], metaclass=ABCMeta):
    """
    Tests for `setup_irods`.
    """
    def test_setup(self):
        icommands_location, service, icommands_controller, icat_controller = setup_irods(self.get_type_to_test())
        self.assertIsInstance(icommands_controller, IrodsBaseExecutablesController)
        self.assertIsInstance(icat_controller, self.get_type_to_test())
        setup_helper = IrodsSetupHelper(icommands_location)
        self.assertEqual(service.version, setup_helper.get_icat_version())


# Setup tests
globals().update(create_tests(_TestSetupIrods, get_classes_to_test(irods_service_controllers, IrodsServiceController),
                              lambda superclass, test_type: "TestSetupWith%s"
                                                            % extract_version_number(test_type.__name__).replace(".", "_")))

# Fix for stupidity of test runners
del _TestSetupIrods, TestUsingType, create_tests, get_classes_to_test

if __name__ == "__main__":
    unittest.main()



import json
import unittest
from abc import ABCMeta

import requests

from useintest.modules.gitlab.gitlab import GitLabServiceController, gitlab_service_controllers
from useintest.services.models import DockerisedServiceWithUsers
from testhelpers import TypeUsedInTest, create_tests, get_classes_to_test
from useintest.tests.services.common import TestServiceControllerSubclass


class _TestGitLabBaseServiceController(
    TestServiceControllerSubclass[TypeUsedInTest, DockerisedServiceWithUsers], metaclass=ABCMeta):
    """
    Tests for `GitLabBaseServiceController`.
    """
    def test_start(self):
        service = self._start_service()
        response = requests.post(f"http://localhost:{service.ports[80]}/api/v3/session", data={
            "login": service.root_user.username, "password": service.root_user.password})
        print(response.text)
        response_payload = json.loads(response.text)
        self.assertEqual(service.root_user.username, response_payload["username"])


# Setup tests
globals().update(create_tests(_TestGitLabBaseServiceController, get_classes_to_test(
    gitlab_service_controllers, GitLabServiceController)))

# Fix for stupidity of test runners
del _TestGitLabBaseServiceController, TestServiceControllerSubclass, create_tests, get_classes_to_test

if __name__ == "__main__":
    unittest.main()



import os

from temphelpers import TempManager
from typing import Tuple, Type

from useintest.common import MOUNTABLE_TEMP_DIRECTORY
from useintest.modules.irods.executables import IrodsExecutablesController, \
    irods_executables_controllers_and_versions
from useintest.modules.irods.models import IrodsDockerisedService
from useintest.modules.irods.services import IrodsServiceController, IrodsBaseServiceController

_temp_manager = TempManager()


def setup_irods(irods_service_controller: Type[IrodsBaseServiceController]=IrodsServiceController) \
        -> Tuple[str, IrodsDockerisedService, IrodsExecutablesController, IrodsBaseServiceController]:
    """
    Sets up an iRODS server and the icommands needed to access the server from the local machine.
    :param irods_service_controller:
    :return: tuple where the first item is the location of the icommands on the host, the second is the iCAT service,
    the thirds is the controller for the icommand executables and the last is the controller for the iCAT service
    """
    global _temp_manager

    # Setup iRODS server
    icat_controller = irods_service_controller()
    service = icat_controller.start_service()

    # Write iRODS connection settings for the server
    settings_directory = _temp_manager.create_temp_directory(dir=MOUNTABLE_TEMP_DIRECTORY)
    config_file = os.path.join(settings_directory, icat_controller.config_file_name)
    password = irods_service_controller.write_connection_settings(config_file, service)

    # Setup iRODS executables
    ExecutablesController = irods_executables_controllers_and_versions[service.version]
    icommands_controller = ExecutablesController(service.name, settings_directory)
    icommands_location = icommands_controller.write_executables_and_authenticate(password)

    return icommands_location, service, icommands_controller, icat_controller



import unittest
from abc import ABCMeta

from testhelpers import TypeUsedInTest, get_classes_to_test, create_tests

from useintest.modules.irods.setup_irods import setup_irods
from useintest.modules.irods.helpers import IrodsSetupHelper
from useintest.modules.irods.services import irods_service_controllers, IrodsServiceController
from useintest.services.models import DockerisedServiceWithUsers
from useintest.tests.services.common import TestDockerisedServiceControllerSubclass


class _TestIrodsServiceController(
        TestDockerisedServiceControllerSubclass[TypeUsedInTest, DockerisedServiceWithUsers], metaclass=ABCMeta):
    """
    Tests for iRODS controller.
    """
    def setUp(self):
        super().setUp()
        self.icommands_location, self.service, self.icommands_controller, self.icat_controller = setup_irods(
            irods_service_controller=self.get_type_to_test())
        self.setup_helper = IrodsSetupHelper(self.icommands_location)

    def tearDown(self):
        super().tearDown()
        self.icat_controller.stop_service(self.service)
        self.icommands_controller.tear_down()

    def test_starts(self):
        name, data = "name", "data"
        data_object_path = self.setup_helper.create_data_object(name, data)
        self.assertEqual(data, self.setup_helper.read_data_object(data_object_path))

    def test_starts_correct_version(self):
        setup_helper = IrodsSetupHelper(self.icommands_location)
        self.assertEqual(self.service.version, self.setup_helper.get_icat_version())


# Setup tests
globals().update(create_tests(_TestIrodsServiceController, get_classes_to_test(irods_service_controllers, IrodsServiceController)))

# Fix for stupidity of test runners
del _TestIrodsServiceController, TestDockerisedServiceControllerSubclass, create_tests, get_classes_to_test

if __name__ == "__main__":
    unittest.main()



import unittest

from docker.errors import NotFound

from useintest.common import docker_client
from useintest.services.builders import DockerisedServiceControllerTypeBuilder

NoopServiceController = DockerisedServiceControllerTypeBuilder(
    name="NoopController",
    repository="alpine",
    ports=[],
    tag="3.6",
    additional_run_settings={"entrypoint": "tail", "command": ["-f", "/etc/hosts"]},
    start_detector=lambda log_line: log_line.strip() != "",
).build()


class TestDockerisedServiceController(unittest.TestCase):
    """
    Tests for `DockerisedServiceController`.
    """
    def setUp(self):
        self._service_controller = NoopServiceController()

    def test_context_manager(self):
        with self._service_controller.start_service() as service:
            self.assertEqual("running", service.container.status)
            container_id = service.container_id
        self.assertIsNone(service.container)
        self.assertRaises(NotFound, docker_client.containers.get, container_id)

    def test_context_manager_exit_when_service_stopped(self):
        with self._service_controller.start_service() as service:
            self._service_controller.stop_service(service)
            self.assertIsNone(service.container)




import unittest
import requests
from abc import ABCMeta

from useintest.modules.bissell.bissell import LatestBissellDockerisedServiceController, BissellServiceController, \
    bissell_service_controllers
from useintest.services.models import DockerisedServiceWithUsers
from testhelpers import TypeUsedInTest, create_tests, get_classes_to_test
from useintest.tests.services.common import TestDockerisedServiceControllerSubclass


class _TestBissellDockerisedServiceController(
        TestDockerisedServiceControllerSubclass[TypeUsedInTest, DockerisedServiceWithUsers], metaclass=ABCMeta):
    """
    Tests for Mongo service controllers.
    """
    def test_start(self):
        service = self._start_service()
        response = requests.head(f"http://{service.host}:{service.port}")
        self.assertEqual(401, response.status_code)


# Setup tests
CLASSES_TO_TEST = {LatestBissellDockerisedServiceController}
globals().update(create_tests(_TestBissellDockerisedServiceController, get_classes_to_test(bissell_service_controllers, BissellServiceController)))


# Fix for stupidity of test runners
del _TestBissellDockerisedServiceController, TestDockerisedServiceControllerSubclass, create_tests, get_classes_to_test


if __name__ == "__main__":
    unittest.main()



from useintest.modules.irods.executables import IrodsBaseExecutablesController, Irods3_3_1ExecutablesController,\
    Irods4_1_8ExecutablesController, Irods4_1_9ExecutablesController, Irods4_1_10ExecutablesController, \
    IrodsExecutablesController, irods_executables_controllers_and_versions, irods_executables_controllers
from useintest.modules.irods.helpers import AccessLevel, IrodsSetupHelper
from useintest.modules.irods.models import IrodsResource, IrodsUser, IrodsDockerisedService
from useintest.modules.irods.setup_irods import setup_irods



import unittest

import os

from useintest.modules.gitlab import SshKey


class TestSshKey(unittest.TestCase):
    """
    Tests for `SshKey`.
    """
    def test_enter_exit_when_no_file_used(self):
        with SshKey():
            pass

    def test_private_key_file(self):
        with SshKey() as ssh_key:
            location = ssh_key.private_key_file
            with open(location, "r") as file:
                self.assertEqual(ssh_key.private_key, file.read())
        self.assertFalse(os.path.exists(location))

    def test_public_key_file(self):
        with SshKey() as ssh_key:
            location = ssh_key.public_key_file
            with open(location, "r") as file:
                self.assertEqual(ssh_key.public_key, file.read())
        self.assertFalse(os.path.exists(location))

    def test_all_key_files(self):
        with SshKey() as ssh_key:
            private_key_file, public_key_file = ssh_key.private_key_file, ssh_key.public_key_file
        self.assertFalse(os.path.exists(private_key_file))
        self.assertFalse(os.path.exists(public_key_file))


if __name__ == "__main__":
    unittest.main()



from abc import ABCMeta
from typing import Type, TypeVar

from useintest.services.controllers import DockerisedServiceController, ServiceType, DockerisedServiceType
from useintest.services.models import DockerisedService, Service

DockerControllerType = TypeVar("DockerControllerType", bound=DockerisedServiceController)


class ServiceControllerTypeBuilder(metaclass=ABCMeta):
    """
    Builder for controllers with particular setups (e.g. repositories and tags).
    """
    def __init__(self, name: str, *args, superclass: Type[DockerControllerType],
                 service_model: Type[ServiceType]=Service, **kwargs):
        """
        Constructor.
        :param name: name of the type that is to be created
        :param args: any arguments that are to be automatically given to the superclass constructor
        :param superclass: the superclass of the type that is to be created
        :param service_model: model of the service the controller type controls
        :param kwargs: any named arguments that are to be automatically given to the superclass constructor
        """
        self.name = name
        self.superclass = superclass
        self.service_model = service_model
        self.args = args
        self.kwargs = kwargs

    def build(self) -> Type[DockerControllerType]:
        """
        Builds the new controller type.
        :return: the new controller type
        """
        def init(controller_self, *args, **kwargs):
            super(type(controller_self), controller_self).__init__(
                self.service_model, *self.args, *args, **{**self.kwargs, **kwargs})

        return type(
            self.name,
            (self.superclass[self.service_model], ),
            {"__init__": init}
        )


class DockerisedServiceControllerTypeBuilder(ServiceControllerTypeBuilder):
    """
    Builder for Docker controllers with particular setups (e.g. repositories and tags).
    """
    def __init__(self, name: str, *args, superclass: Type[DockerisedServiceController]=DockerisedServiceController,
                 service_model: Type[DockerisedServiceType]=DockerisedService, **kwargs):
        super().__init__(name, *args, superclass=superclass, service_model=service_model, **kwargs)



from useintest.common import UseInTestError


class ServiceStartError(UseInTestError):
    """
    Exception for when a service has failed to start.
    """


class TransientServiceStartError(ServiceStartError):
    """
    Exception for when a service has failed to start due to a (suspected) transient issue.
    """


class PersistentServiceStartError(ServiceStartError):
    """
    Exception for when a service has failed to start due to a (suspected) persistent issue.
    """


class UnexpectedNumberOfPortsError(UseInTestError):
    """
    Exception for when the number of ports is not as expected.
    """



from typing import List, Any

import semantic_version

from useintest.common import UseInTestModel
from useintest.services.models import DockerisedService

# Import from semantic version library
Version = semantic_version.Version


class IrodsResource(UseInTestModel):
    """
    Model of a iRODS server resource.
    """
    def __init__(self, name: str, host: str, location: str):
        self.name = name
        self.host = host
        self.location = location


# TODO: Extend User in /common
class IrodsUser(UseInTestModel):
    """
    Model of an iRODS user.
    """
    def __init__(self, username: str, zone: str, password: str=None, admin: bool=False):
        super().__init__()
        self.username = username
        self.password = password
        self.zone = zone
        self.admin = admin

    def __eq__(self, other: Any) -> bool:
        return type(other) == type(self) \
               and other.username == self.username \
               and other.password == self.password \
               and other.zone == self.zone \
               and other.admin == self.admin


# TODO: Extend ServiceWithUsers in /common
class IrodsDockerisedService(DockerisedService):
    """
    Model of a iRODS service running in Docker.
    """
    def __init__(self):
        super().__init__()
        self.users: List[IrodsUser] = []
        self.version: Version = None



from useintest.modules.mongo.mongo import mongo_service_controllers, MongoServiceController, Mongo3ServiceController, \
    MongoLatestDockerisedServiceController, Mongo3DockerisedServiceController, common_setup



import logging
import os
from typing import List, Any, Set

from docker.errors import ImageNotFound

from useintest.common import docker_client

CLI_ARGUMENTS = "\"$@\""

SHEBANG = "#!/usr/bin/env bash"
FAIL_SETTINGS = "set -eu -o pipefail"


def write_commands(location: str, commands: str):
    """
    Writes the commands to a file at the given location. Will overwrite any pre-existing file.
    :param location: the location to write the commands to
    :param commands: the commands to write
    """
    with open(location, "w") as file:
        file.write("%s\n" % SHEBANG)
        file.write("%s\n" % FAIL_SETTINGS)
        file.write(commands)
    os.chmod(location, 0o700)


def pull_docker_image(image: str, tag: str=None):
    """
    TODO
    :param image:
    :param tag:
    :return:
    """
    # Ensure the image with the real binaries have been pulled to stop it polluting the output
    if ":" in image:
        if tag is not None:
            raise ValueError("Cannot specify tag when tag has been passed in the image parameter")
        repository, tag = image.split(":")
    else:
        repository, tag = image, tag

    try:
        docker_client.images.get(f"{repository}:{tag}")
    except ImageNotFound:
        pull_stream = docker_client.api.pull(repository, tag=tag, stream=True)
        for line in pull_stream:
            # TODO: Remove logging to root logger
            logging.debug(line)


# TODO: Test this
def get_all_path_like_arguments_for_mounting(arguments: List[Any], allow_relative_paths: bool=True) -> Set[str]:
    """
    TODO
    :param arguments:
    :param allow_relative_paths:
    :return:
    """
    mounts = set()  # type: Set[str]
    for argument in arguments:
        if allow_relative_paths:
            argument = os.path.abspath(argument)
        if argument.startswith(os.path.sep):
            mounts.add(os.path.dirname(argument))
    return mounts



import os
import tempfile
import unittest
from uuid import uuid4

from temphelpers import TempManager
from typing import List, Tuple

from useintest.executables.builders import CommandsBuilder, MountedArgumentParserBuilder
from useintest.executables.common import write_commands
from useintest.executables.controllers import ExecutablesController
from useintest.executables.models import Executable
from useintest.tests.executables.common import get_builder_for_commands_to_run_persistent_ubuntu, run, \
    UBUNTU_IMAGE_TO_TEST_WITH
from useintest.tests.common import MOUNTABLE_TEMP_CREATION_KWARGS
from useintest.common import MOUNTABLE_TEMP_DIRECTORY

_CONTENT = "Hello World!"
_CAT_MOUNTED_ARGUMENT_PARSER = MountedArgumentParserBuilder(
    positional_arguments=MountedArgumentParserBuilder.ALL_POSITIONAL_ARGUMENTS).build()
_TOUCH_MOUNTED_ARGUMENT_PARSER = MountedArgumentParserBuilder(
    named_arguments={"-r"}, positional_arguments=MountedArgumentParserBuilder.ALL_POSITIONAL_ARGUMENTS).build()


class TestExecutablesController(unittest.TestCase):
    """
    Tests for ExecutablesController.
    """
    def setUp(self):
        self._temp_manager = TempManager(MOUNTABLE_TEMP_CREATION_KWARGS, MOUNTABLE_TEMP_CREATION_KWARGS)
        self.controller = ExecutablesController()
        self.persistent_run_controller = ExecutablesController(get_builder_for_commands_to_run_persistent_ubuntu())

    def tearDown(self):
        self.controller.tear_down()
        self.persistent_run_controller.tear_down()
        self._temp_manager.tear_down()

    def test_create_executable_commands_with_no_running_container(self):
        commands_builder = CommandsBuilder("echo", image=UBUNTU_IMAGE_TO_TEST_WITH, executable_arguments=[_CONTENT])
        commands = self.controller.create_executable_commands(Executable(commands_builder, False))
        out, error = self._run_commands(commands)
        self.assertEqual(_CONTENT, out)

    def test_create_executable_commands_with_positional_parameter_needing_mounting_for_read(self):
        _, read_file = self._temp_manager.create_temp_file()
        with open(read_file, "w") as file:
            file.write(_CONTENT)

        commands_builder = CommandsBuilder(
            "cat", image=UBUNTU_IMAGE_TO_TEST_WITH, get_path_arguments_to_mount=_CAT_MOUNTED_ARGUMENT_PARSER)
        commands = self.controller.create_executable_commands(Executable(commands_builder, False))
        out, error = self._run_commands(commands, [read_file])
        self.assertEqual(_CONTENT, out)

    def test_create_executable_commands_with_positional_parameter_needing_mounting_for_write(self):
        with tempfile.TemporaryDirectory(dir=MOUNTABLE_TEMP_DIRECTORY) as temp_directory:
            file_locations = [os.path.join(temp_directory, str(uuid4())) for _ in range(5)]
            commands_builder = CommandsBuilder(
                "touch", image=UBUNTU_IMAGE_TO_TEST_WITH, get_path_arguments_to_mount=_TOUCH_MOUNTED_ARGUMENT_PARSER)
            commands = self.controller.create_executable_commands(Executable(commands_builder, False))
            self._run_commands(commands, file_locations)

            for location in file_locations:
                self.assertTrue(os.path.exists(location))

    def test_create_executable_commands_with_named_parameters_needing_mounting(self):
        with tempfile.TemporaryDirectory(dir=MOUNTABLE_TEMP_DIRECTORY) as temp_directory:
            commands_builder = CommandsBuilder(
                "touch", image=UBUNTU_IMAGE_TO_TEST_WITH, get_path_arguments_to_mount=_TOUCH_MOUNTED_ARGUMENT_PARSER,
                mounts={temp_directory: temp_directory})
            commands = self.persistent_run_controller.create_executable_commands(Executable(commands_builder, False))

            test_file = os.path.join(temp_directory, "test-file")
            _, reference = self._temp_manager.create_temp_file()
            os.utime(reference, (0, 0))
            self._run_commands(commands, ["-r", reference, test_file])
            self.assertEqual(0.0, os.path.getmtime(test_file))

    def test_create_simple_executable_commands_that_raises_exception(self):
        commands = self.persistent_run_controller.create_simple_executable_commands("cat", "/does-not-exist")
        out, error = self._run_commands(commands, raise_if_stderr=False)
        self.assertNotEqual("", error)
        self.assertEqual("", out)

    def test_create_simple_executable_commands_execute_in_same_container(self):
        test_file_location = "/test-file"
        self._run_commands(self.persistent_run_controller.create_simple_executable_commands("touch", test_file_location))
        out, error = self._run_commands(self.persistent_run_controller.create_simple_executable_commands("ls", test_file_location))
        self.assertEqual(test_file_location, out)

    def test_create_simple_executable_commands_without_parameters(self):
        commands = self.persistent_run_controller.create_simple_executable_commands("cat", "/etc/lsb-release")
        out, error = self._run_commands(commands)
        key_values = {line.split("=")[0]: line.split("=")[1] for line in out.split("\n")}
        self.assertEqual("Ubuntu", key_values["DISTRIB_ID"])

    def test_create_simple_executable_commands_with_parameters(self):
        commands = self.persistent_run_controller.create_simple_executable_commands("echo")
        out, error = self._run_commands(commands, [_CONTENT])
        self.assertEqual(out, _CONTENT)

    def _run_commands(self, commands: str, arguments: List=None, raise_if_stderr: bool=True) -> Tuple[str, str]:
        """
        Saves the given commands as an executable and runs it with the given arguments.
        :param commands: the commands to execute
        :param arguments: the arguments to be passed to the executable
        :param raise_if_stderr: whether the test should raise an exception if anything is written to standard error
        :return: tuple where the first element is what was written to standard out and the second is what was written to
        standard error
        """
        if arguments is None:
            arguments = []

        file_handle, location = self._temp_manager.create_temp_file()
        os.close(file_handle)
        write_commands(location, commands)

        return run([location] + arguments, raise_if_stderr)


if __name__ == "__main__":
    unittest.main()



import os
import unittest
from abc import ABCMeta

from testhelpers import TestUsingType, TypeUsedInTest, create_tests

from useintest.modules.irods.models import Version
from useintest.modules.samtools import samtools_executable_controllers
from useintest.tests.executables.common import run

EXAMPLE_BAM = os.path.join(os.path.dirname(os.path.realpath(__file__)), "example.bam")
assert os.path.exists(EXAMPLE_BAM)


class _TestSamtoolsExecutablesController(TestUsingType[TypeUsedInTest], metaclass=ABCMeta):
    """
    Tests for executables controllers for Samtools.
    """

    def setUp(self):
        self.controller = self.get_type_to_test()()
        self.version = Version(self.get_type_to_test().__name__.replace("Samtools", "")
                               .replace("ExecutablesController", "").replace("_", "."))
        self.executables_location = self.controller.write_executables()
        self.samtools_location = os.path.join(self.executables_location, "samtools")

    def tearDown(self):
        self.controller.tear_down()

    def test_help(self):
        out, _ = run([self.samtools_location, "--help"])
        self.assertIn("Version: %s" % self.version, out)

    def test_view(self):
        original_sam_contents, _ = run([self.samtools_location, "view", "-O", "SAM", EXAMPLE_BAM], decode_output_to=None)
        sam_contents, _ = run([self.samtools_location, "view", "-O", "SAM", EXAMPLE_BAM], decode_output_to=None)
        self.assertEqual(original_sam_contents, sam_contents)


# Setup tests
globals().update(create_tests(_TestSamtoolsExecutablesController, samtools_executable_controllers))

# Fix for stupidity of test runners
del _TestSamtoolsExecutablesController, TestUsingType, create_tests

if __name__ == "__main__":
    unittest.main()



from enum import Enum, unique
from pathlib import Path
from tempfile import NamedTemporaryFile
from threading import Lock

import os

from typing import Callable, Dict

from useintest.common import MissingDependencyError

_UTF8_ENCODING = "utf-8"


@unique
class _KeyType(Enum):
    """"
    Types of keys.
    """
    PRIVATE = "private"
    PUBLIC = "public"


class SshKey:
    """
    Wrapper to help use an SSH key.
    """
    @property
    def private_key(self) -> str:
        return self.lazy_get_value(_KeyType.PRIVATE, lambda: self._key.exportKey().decode(_UTF8_ENCODING))

    @property
    def public_key(self) -> str:
        return self.lazy_get_value(_KeyType.PUBLIC, lambda: self._key.publickey().exportKey("OpenSSH").decode(
            _UTF8_ENCODING))

    @property
    def private_key_file(self) -> str:
        return self._lazy_get_file(_KeyType.PRIVATE, lambda: self.private_key)

    @property
    def public_key_file(self) -> str:
        return self._lazy_get_file(_KeyType.PUBLIC, lambda: self.public_key)

    def __init__(self, key_length: int=2048):
        """
        Constructor.
        :param key_length: length of the RSA key to generate.
        """
        try:
            from Cryptodome.PublicKey import RSA
        except ImportError as e:
            raise MissingDependencyError("pycryptodomex") from e

        self._key = RSA.generate(key_length)
        self._values: Dict[_KeyType, str] = {}
        self._file_locations: Dict[_KeyType: str] = {}
        self._locks = {
            _KeyType.PRIVATE: Lock(),
            _KeyType.PUBLIC: Lock()
        }

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.tear_down()

    def tear_down(self):
        """
        Removes temp files managed by this instance.
        """
        for key_type, lock in self._locks.items():
            with lock:
                if key_type in self._file_locations:
                    try:
                        os.remove(self._file_locations[key_type])
                    except OSError:
                        pass

    def lazy_get_value(self, key_type: _KeyType, data_source: Callable[[], str]):
        """
        Lazily get the value associated to the given key type, generating it from the given data source and saving it to
        the value cache if it has not been set.
        :param key_type: the type of the key
        :param data_source: the source to generate the value from if not generated
        :return: the value associated to the given key
        """
        if key_type not in self._values:
            self._values[key_type] = data_source()
        return self._values[key_type]

    def _lazy_get_file(self, key_type: _KeyType, data_source: Callable[[], str]) -> str:
        """
        Lazily gets the file associated to the given key type, generating it from the given data source and saving it to
        the file cache if it has not been set.
        :param key_type: the type of the key
        :param data_source: the source to generate the file from if not generated
        :return: the file associated to the given key
        """
        if key_type not in self._file_locations:
            with self._locks[key_type]:
                if key_type not in self._file_locations:
                    temp_file = NamedTemporaryFile().name
                    Path(temp_file).touch()
                    os.chmod(temp_file, 0o600)
                    with open(temp_file, "w") as file:
                        file.write(data_source())
                    self._file_locations[key_type] = temp_file
        return self._file_locations[key_type]



import json
import logging
import math
import os
from abc import abstractmethod, ABCMeta
from typing import List, Type, Callable, Sequence

from useintest.modules.irods.models import IrodsUser, IrodsDockerisedService, Version
from useintest.services.controllers import DockerisedServiceController

_logger = logging.getLogger(__name__)

_DOCKER_REPOSITORY = "mercury/icat"


class IrodsBaseServiceController(DockerisedServiceController, metaclass=ABCMeta):
    """
    TODO
    """
    @staticmethod
    @abstractmethod
    def write_connection_settings(file_location: str, service: IrodsDockerisedService) -> str:
        """
        Writes the connection settings for the given iRODS service to the given location.
        :param file_location: the location to write the settings to (file should not already exist)
        :param service: the Dockerized iRODS service
        """

    @staticmethod
    def _persistent_error_detector(log_line: str) -> bool:
        """
        TODO
        :param log_line:
        :return:
        """
        return "No space left on device" in log_line

    def __init__(self, version: Version, users: Sequence[IrodsUser], config_file_name: str,
                 repository: str, tag: str, ports: List[int], start_detector: Callable[[str], bool], **kwargs):
        """
        Constructor.
        :param version:
        :param users:
        :param config_file_name:
        :param repository:
        :param tag:
        :param ports:
        :param start_detector:
        :param kwargs:
        """
        super().__init__(IrodsDockerisedService, repository, tag, ports, start_detector, **kwargs)
        self.config_file_name = config_file_name
        self._version = version
        self._users = users

    def start_service(self) -> IrodsDockerisedService:
        service = super().start_service()
        service.users = self._users
        service.version = self._version
        return service


class Irods4ServiceController(IrodsBaseServiceController, metaclass=ABCMeta):
    """
    iRODS 4 service controller.
    """
    _PORT = 1247
    _CONFIG_FILE_NAME = "irods_environment.json"

    _HOST_PARAMETER_NAME = "irods_host"
    _PORT_PARAMETER_NAME = "irods_port"
    _USERNAME_PARAMETER_NAME = "irods_user_name"
    _ZONE_PARAMETER_NAME = "irods_zone_name"

    _USERS = [IrodsUser("rods", "testZone", "irods123", admin=True)]

    # TODO: These connection settings will not work with port-mapping to localhost
    @staticmethod
    def write_connection_settings(file_location: str, service: IrodsDockerisedService) -> str:
        if os.path.isfile(file_location):
            raise ValueError("Settings cannot be written to a file that already exists (%s)" % file_location)

        user = service.users[0]
        config = {
            Irods4ServiceController._USERNAME_PARAMETER_NAME: user.username,
            Irods4ServiceController._HOST_PARAMETER_NAME: service.name,
            Irods4ServiceController._PORT_PARAMETER_NAME: Irods4ServiceController._PORT,
            Irods4ServiceController._ZONE_PARAMETER_NAME: user.zone
        }
        config_as_json = json.dumps(config)
        logging.debug("Writing iRODS connection config to: %s" % file_location)
        with open(file_location, "w") as settings_file:
            settings_file.write(config_as_json)

        return user.password

    @staticmethod
    def _start_detector(log_line: str) -> bool:
        """
        TODO
        :param log_line:
        :return:
        """
        return "iRODS server started successfully!" in log_line

    @staticmethod
    def _transient_error_detector(log_line: str) -> bool:
        """
        TODO
        :param log_line:
        :return:
        """
        # Note: iRODS schema validation has been observed to randomly fail before, raising `RuntimeError`
        return "iRODS server failed to start." in log_line or "RuntimeError:" in log_line

    def __init__(self, docker_repository: str, docker_tag: str, start_timeout: int=math.inf, start_tries: int=10,
                 version: Version=None):
        """
        Constructor.
        :param docker_repository: name of the Docker repository
        :param docker_tag: the Docker tag of the iRODS 4 image
        :param start_timeout: see `ContainerisedServiceController.__init__`
        :param start_tries: see `ContainerisedServiceController.__init__`
        :param version: exact version of the iRODS 4 sever (will use `docker_tag` if not supplied)
        """
        version = version if version is not None else Version(docker_tag)
        super().__init__(version, Irods4ServiceController._USERS, Irods4ServiceController._CONFIG_FILE_NAME,
                         docker_repository, docker_tag, [Irods4ServiceController._PORT],
                         Irods4ServiceController._start_detector,
                         transient_error_detector=Irods4ServiceController._transient_error_detector,
                         persistent_error_detector=IrodsBaseServiceController._persistent_error_detector,
                         start_timeout=start_timeout, start_tries=start_tries)


# TODO: Why not use DockerisedServiceControllerTypeBuilder?
def build_irods_service_controller_type(docker_repository: str, docker_tag: str, superclass: type) \
        -> Type[IrodsBaseServiceController]:
    """
    Builds a controller for an iRODS server that runs in containers of on the given Docker image.
    :param docker_repository: name of the Docker repository
    :param docker_tag: the Docker tag of the image in the Docker repository
    :param superclass: the superclass of the service controller
    :return: the build service controller for the given image
    """
    def init(self: superclass, *args, **kwargs):
        super(type(self), self).__init__(docker_repository, docker_tag, *args, **kwargs)

    return type(
        "Irods%sServiceController" % docker_tag.replace(".", "_"),
        (superclass,),
        {"__init__": init}
    )


# Concrete service controller definitions
Irods4_1_10ServiceController = build_irods_service_controller_type(_DOCKER_REPOSITORY, "4.1.10", Irods4ServiceController)
IrodsServiceController = Irods4_1_10ServiceController

irods_service_controllers = {Irods4_1_10ServiceController}



from useintest.services.builders import DockerisedServiceControllerTypeBuilder

common_setup = {
    "repository": "mongo",
    "start_detector": lambda log_line: "waiting for connections on port" in log_line,
    "persistent_error_detector": lambda log_line: "error creating journal dir" in log_line
                                                  or "No space left on device" in log_line,
    "ports": [27017]
}

Mongo3DockerisedServiceController = DockerisedServiceControllerTypeBuilder(
    name="Mongo3DockerController",
    tag="3",
    **common_setup).build()   # type: type

MongoLatestDockerisedServiceController = DockerisedServiceControllerTypeBuilder(
    name="MongoLatestDockerController",
    tag="latest",
    **common_setup).build()   # type: type


Mongo3ServiceController = Mongo3DockerisedServiceController
MongoServiceController = MongoLatestDockerisedServiceController

mongo_service_controllers = {Mongo3ServiceController, MongoServiceController}


from useintest.services.builders import DockerisedServiceControllerTypeBuilder

common_setup = {
    "repository": "couchdb",
    "start_detector": lambda log_line: "Apache CouchDB has started" in log_line,
    "persistent_error_detector": lambda log_line: "no space left on device" in log_line,
    "ports": [5984]
}

CouchDB1_6DockerisedServiceController = DockerisedServiceControllerTypeBuilder(
    name="CouchDB1_6DockerisedServiceController",
    tag="1.6",
    **common_setup).build()   # type: type


CouchDB1_6ServiceController = CouchDB1_6DockerisedServiceController
CouchDBServiceController = CouchDB1_6ServiceController

couchdb_service_controllers = {CouchDB1_6ServiceController, CouchDBServiceController}



from time import sleep

import requests

from useintest.services.builders import DockerisedServiceControllerTypeBuilder
from useintest.services.exceptions import TransientServiceStartError
from useintest.services.models import DockerisedService


def startup_monitor(service: DockerisedService):
    """
    Checks whether Bissell has started.

    Discussion on how to do this can be found here: https://github.com/wtsi-hgi/bissell/issues/1.
    :param service: the Bissell service
    :return: `True`
    """
    while True:
        logs = service.container.logs()
        if len(logs) > 0:
            TransientServiceStartError(f"Bissell output detected in logs indicates a startup failure: {logs}")
        try:
            response = requests.head(f"http://{service.host}:{service.port}")
            if response.status_code == 401:
                return True
        except requests.exceptions.ConnectionError:
            pass
        sleep(10)


common_setup = {
    "repository": "mercury/bissell",
    "startup_monitor": startup_monitor,
    "ports": [5000]
}

LatestBissellDockerisedServiceController = DockerisedServiceControllerTypeBuilder(
    name="BissellDockerController",
    tag="latest",
    **common_setup).build()   # type: type

BissellServiceController = LatestBissellDockerisedServiceController

bissell_service_controllers = {BissellServiceController}



from setuptools import setup, find_packages

try:
    from pypandoc import convert
    def read_markdown(file: str) -> str:
        return convert(file, "rst")
except ImportError:
    def read_markdown(file: str) -> str:
        return open(file, "r").read()

setup(
    name="useintest",
    version="5.0.0",
    author="Colin Nolan",
    author_email="colin.nolan@sanger.ac.uk",
    packages=find_packages(exclude=["tests"]),
    install_requires=[x for x in open("requirements.txt").read().splitlines() if "://" not in x],
    dependency_links=[x for x in open("requirements.txt").read().splitlines() if "://" in x],
    url="https://github.com/wtsi-hgi/useintest",
    license="MIT",
    description="I don't care how it's done, I just want to use it in my tests!",
    long_description=read_markdown("README.md"),
    test_suite="useintest.tests",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License"
    ]
)



import logging
import os
import tarfile
from abc import ABCMeta
from io import BytesIO
from pathlib import PurePosixPath
from time import sleep
from typing import Generic

from useintest.services.builders import DockerisedServiceControllerTypeBuilder
from useintest.services.controllers import DockerisedServiceController, DockerisedServiceWithUsersType
from useintest.services.models import User, DockerisedServiceWithUsers

_CONFIGURATION_HOST_LOCATION = os.path.join(os.path.dirname(os.path.abspath(__file__)), "_resources/data")
_CONFIGURATION_DOCKER_LOCATION = PurePosixPath("/data")

_ROOT_USERNAME = "root"
_ROOT_PASSWORD = "root"


class GogsBaseServiceController(Generic[DockerisedServiceWithUsersType],
                                DockerisedServiceController[DockerisedServiceWithUsersType], metaclass=ABCMeta):
    """
    Base class for Gogs service controllers.
    """
    def start_service(self) -> DockerisedServiceWithUsersType:
        service = super().start_service()
        container = service.container

        # Painful conversion required due to limitation of the Docker client:
        # https://github.com/docker/docker-py/issues/1027#issuecomment-299654299
        archive_container = BytesIO()
        with tarfile.open(mode="w", fileobj=archive_container) as archive:
            archive.add(_CONFIGURATION_HOST_LOCATION, arcname=_CONFIGURATION_DOCKER_LOCATION.name)
        archive_as_bytes = bytes(archive_container.getbuffer())

        # Add configuration
        container.put_archive(_CONFIGURATION_DOCKER_LOCATION.parent.as_posix(), archive_as_bytes)
        service.root_user = User(_ROOT_USERNAME, _ROOT_PASSWORD)

        # Start gogs
        _, output_generator = container.exec_run(["/app/gogs/docker/start.sh"], stream=True)
        for line in output_generator:
            logging.debug(line)
            if "Listen: http://0.0.0.0:3000" in str(line):
                break

        return service


# Employing hacky way of getting run sleeping forever with the detector
common_setup = {
    "superclass": GogsBaseServiceController,
    "service_model": DockerisedServiceWithUsers,
    "repository": "gogs/gogs",
    "start_detector": lambda log_line: "127.0.0.1\tlocalhost" in log_line,
    "transient_error_detector": lambda log_line: "the container has stopped" in log_line,
    "ports": [3000],
    "additional_run_settings": {"entrypoint": "tail", "command": ["-f", "/etc/hosts"]}
}

Gogs0_11_4ServiceController = DockerisedServiceControllerTypeBuilder(
    name="Gogs0_11_4ServiceController",
    tag="0.11.4",
    **common_setup).build()

GogsServiceController = Gogs0_11_4ServiceController

gogs_service_controllers = {Gogs0_11_4ServiceController}



import unittest
from abc import ABCMeta

from consul import Consul

from useintest.modules.consul.consul import ConsulServiceController, consul_service_controllers, \
    ConsulDockerisedService
from useintest.services.models import DockerisedServiceWithUsers
from testhelpers import TypeUsedInTest, create_tests, get_classes_to_test
from useintest.tests.services.common import TestServiceControllerSubclass

_TEST_KEY = "hello"
_TEST_VALUE = "world"


class _TestConsulServiceController(
        TestServiceControllerSubclass[TypeUsedInTest, DockerisedServiceWithUsers], metaclass=ABCMeta):
    """
    Tests for `ConsulServiceController`.
    """
    def test_start(self):
        service = self._start_service()     # type: ConsulDockerisedService
        self.assertIsInstance(service, ConsulDockerisedService)
        consul_client = service.create_consul_client()
        self._test_client(consul_client)

    def test_setup_environment(self):
        service = self._start_service()     # type: ConsulDockerisedService
        service.setup_environment()
        consul_client = Consul()
        self._test_client(consul_client)

    def _test_client(self, consul_client: Consul):
        """
        Tests that the given Consul client connects to Consul
        :param consul_client: the Consul client to test
        """
        consul_client.kv.put(_TEST_KEY, _TEST_VALUE)
        self.assertEqual(_TEST_VALUE, consul_client.kv.get(_TEST_KEY)[1]["Value"].decode("utf-8"))


# Setup tests
globals().update(create_tests(_TestConsulServiceController, get_classes_to_test(
    consul_service_controllers, ConsulServiceController)))

# Fix for stupidity of test runners
del _TestConsulServiceController, TestServiceControllerSubclass, create_tests, get_classes_to_test

if __name__ == "__main__":
    unittest.main()



import unittest
from abc import ABCMeta

from gogs_client import UsernamePassword, GogsApi
from testhelpers import TypeUsedInTest, create_tests, get_classes_to_test

from useintest.modules.gogs.gogs import gogs_service_controllers, GogsServiceController
from useintest.services.models import DockerisedServiceWithUsers
from useintest.tests.services.common import TestServiceControllerSubclass

_REPO_NAME = "test"


class _TestGogsBaseServiceController(
    TestServiceControllerSubclass[TypeUsedInTest, DockerisedServiceWithUsers], metaclass=ABCMeta):
    """
    Tests for `GogsBaseServiceController`.
    """
    def test_start(self):
        service = self._start_service()

        authentication = UsernamePassword(service.root_user.username, service.root_user.password)
        gogs_connection = GogsApi(f"http://{service.host}:{service.ports[3000]}")
        gogs_connection.create_repo(authentication, _REPO_NAME)

        self.assertEqual(f"{authentication.username}/{_REPO_NAME}",
                          gogs_connection.get_repo(authentication, authentication.username, _REPO_NAME).full_name)


# Setup tests
globals().update(create_tests(_TestGogsBaseServiceController, get_classes_to_test(
    gogs_service_controllers, GogsServiceController)))

# Fix for stupidity of test runners
del _TestGogsBaseServiceController, TestServiceControllerSubclass, create_tests, get_classes_to_test

if __name__ == "__main__":
    unittest.main()



from useintest.common import UseInTestModel
from useintest.executables.builders import CommandsBuilder


class Executable(UseInTestModel):
    """
    Model of an executable.
    """
    def __init__(self, commands_builder: CommandsBuilder, uses_running_container: bool=True):
        self.commands_builder = commands_builder
        self.uses_running_container = uses_running_container



import atexit
import logging
import math
import socket
from abc import ABCMeta, abstractmethod
from inspect import signature
from typing import Dict, Iterator, List, Callable, TypeVar, Generic, Type
from uuid import uuid4

from docker.errors import NotFound
from stopit import ThreadingTimeout, TimeoutException

from useintest.executables.common import pull_docker_image
from useintest.common import docker_client
from useintest.services.exceptions import ServiceStartError, TransientServiceStartError, \
    PersistentServiceStartError
from useintest.services.models import Service, DockerisedService, DockerisedServiceWithUsers

_logger = logging.getLogger(__name__)

ServiceType = TypeVar("ServiceType", bound=Service)
DockerisedServiceType = TypeVar("DockerisedServiceType", bound=DockerisedService)
DockerisedServiceWithUsersType = TypeVar("DockerisedServiceWithUsersType", bound=DockerisedServiceWithUsers)


def _get_open_port() -> int:
    """
    Gets a PORT that will (probably) be available on the machine.
    It is possible that in-between the time in which the open PORT of found and when it is used, another process may
    bind to it instead.
    :return: the (probably) available PORT
    """
    free_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    free_socket.bind(("", 0))
    free_socket.listen(1)
    port = free_socket.getsockname()[1]
    free_socket.close()
    return port


class ServiceController(Generic[ServiceType], metaclass=ABCMeta):
    """
    Service controller.
    """
    def __init__(self, service_model: Type[ServiceType]):
        """
        Constructor.
        :param service_model: the type of model for the service this controller handles
        """
        # XXX: It would nice to do `ServiceModel()` but I don't think this is possible in Python
        self._service_model = service_model

    @abstractmethod
    def start_service(self) -> ServiceType:
        """
        Starts a service.
        :raises ServiceStartException: service could not be started (see logs for more information)
        :return: model of the started service
        """

    @abstractmethod
    def stop_service(self, service: ServiceType):
        """
        Stops the given service.
        :param service: model of the service to stop
        """


class ContainerisedServiceController(Generic[ServiceType], ServiceController[ServiceType], metaclass=ABCMeta):
    """
    Controller of containers running a service brought up for testing.
    """
    @abstractmethod
    def _start(self, service: Service):
        """
        Starts a container.
        :param service: model of the service to start
        """

    @abstractmethod
    def _stop(self, service: Service):
        """
        Stops the given container.
        :param service: model of the service to stop
        """

    @abstractmethod
    def _wait_until_started(self, service: Service) -> bool:
        """
        Blocks until the given container has started.
        :raises ServiceStartException: raised if service cannot be started
        :param service: the service
        :return: `True` if the service has started successfully
        """

    def __init__(self, service_model: Type[ServiceType], start_timeout: float=math.inf, start_tries: int=10,
                 stop_on_exit: bool=True):
        """
        Constructor.
        :param stop_on_exit: see `Container.__init__`
        :param start_timeout: timeout before for container start
        :param start_tries: number of times to try to start the container before giving up (will only try once if a
        `PersistentServiceStartException` is raised
        :param stop_on_exit: whether to stop all started containers on exit
        """
        super().__init__(service_model)
        self.start_timeout = start_timeout
        self.start_tries = start_tries
        self.stop_on_exit = stop_on_exit

    def start_service(self) -> ServiceType:
        service = self._service_model()
        assert service is not None
        if self.stop_on_exit:
            atexit.register(self.stop_service, service)

        started = False
        tries = 0
        while not started:
            if tries >= self.start_tries:
                raise ServiceStartError()
            if tries > 0:
                self._stop(service)
            self._start(service)
            try:
                if self.start_timeout is not math.inf:
                    with ThreadingTimeout(self.start_timeout, swallow_exc=False):
                        started = self._wait_until_started(service)
                else:
                    started = self._wait_until_started(service)
            except TimeoutException as e:
                _logger.warning(e)
            except TransientServiceStartError as e:
                _logger.warning(e)

            tries += 1
        return service

    def stop_service(self, service: ServiceType):
        self._stop(service)


class DockerisedServiceController(
        Generic[DockerisedServiceType], ContainerisedServiceController[DockerisedServiceType], metaclass=ABCMeta):
    """
    Controller of Docker containers running a service brought up for testing.
    """
    @staticmethod
    def _call_detector_with_correct_arguments(detector: Callable, line: str, service: DockerisedServiceType) \
            -> bool:
        """
        Calls the given detector with either line as the only argument or both line and service, depending on the
        detector's signature.
        :param detector: the detector to call
        :param line: the log line to give to the detector
        :param service: the service being started
        :return: the detector's return value
        """
        number_of_parameters = len(signature(detector).parameters)
        if number_of_parameters == 1:
            return detector(line)
        else:
            return detector(line, service)

    def __init__(self, service_model: Type[ServiceType], repository: str, tag: str, ports: List[int],
                 start_detector: Callable[..., bool]=None,
                 persistent_error_detector: Callable[..., bool]=None,
                 transient_error_detector: Callable[..., bool]=None,
                 start_timeout: int=math.inf, start_tries: int=math.inf, additional_run_settings: dict=None,
                 startup_monitor: Callable[[DockerisedServiceType], bool]=None):
        """
        Constructor.
        :param service_model: see `ServiceController.__init__`
        :param repository: the Dockerhub repository of the service to start
        :param tag: the Dockerhub repository tag of the service to start
        :param ports: the ports the service exposes
        :param start_detector: function that detects if the service is ready to use from the logs
        :param persistent_error_detector: function that detects if the service is unable to start
        :param transient_error_detector: function that detects if the service encountered a transient error when
        starting
        :param start_timeout: timeout for starting containers
        :param start_tries: number of times to try starting the containerised service
        :param additional_run_settings: other run settings (see https://docker-py.readthedocs.io/en/1.2.3/api/#create_container)
        :param startup_monitor: override the default, log based code that waits for the service to start, with a custom
        monitor that should return `True` when the service has started or raise an exception if it is not going to start
        """
        super().__init__(service_model, start_timeout, start_tries)
        self.repository = repository
        self.tag = tag
        self.ports = ports
        self.start_detector = start_detector
        self.persistent_error_detector = persistent_error_detector
        self.transient_error_detector = transient_error_detector
        self.run_settings = additional_run_settings if additional_run_settings is not None else {}
        self.startup_monitor = startup_monitor
        self._log_iterator: Dict[Service, Iterator] = dict()

    def _start(self, service: DockerisedServiceType):
        pull_docker_image(self.repository, self.tag)
        image = docker_client.images.pull(self.repository, tag=self.tag)

        service.name = f"{self.repository.split('/')[-1]}-{uuid4()}"
        service.ports = {port: _get_open_port() for port in self.ports}
        service.controller = self

        container = docker_client.containers.create(
            image=image.id,
            name=service.name,
            ports=service.ports,
            detach=True,
            **self.run_settings)
        service.container = container

        container.start()

    def _stop(self, service: DockerisedServiceType):
        if service in self._log_iterator:
            del self._log_iterator[service]
        if service.container:
            try:
                service.container.stop()
                service.container.remove(force=True)
            except NotFound:
                pass

    def _wait_until_started(self, service: DockerisedServiceType) -> bool:
        if self.startup_monitor is not None:
            return self.startup_monitor(service)
        else:
            log_stream = service.container.logs(stream=True)
            for line in log_stream:
                # XXX: Although non-streamed logs are returned as a string, the generator returns bytes!?
                # http://docker-py.readthedocs.io/en/stable/containers.html#docker.models.containers.Container.logs
                line = line.decode("utf-8")
                logging.debug(line)

                if self.persistent_error_detector is not None \
                        and self._call_detector_with_correct_arguments(self.persistent_error_detector, line, service):
                    raise PersistentServiceStartError(line)
                elif self.transient_error_detector is not None \
                        and self._call_detector_with_correct_arguments(self.transient_error_detector, line, service):
                    raise TransientServiceStartError(line)
                elif self._call_detector_with_correct_arguments(self.start_detector, line, service):
                    return True

            assert len(docker_client.containers.list(filters=dict(name=service.name))) == 0
            logs = service.container.logs()
            raise TransientServiceStartError(
                f"No error detected in logs but the container has stopped. Log dump: {logs}")



from abc import ABCMeta

from useintest.services.models import User, DockerisedServiceWithUsers
from useintest.services.builders import DockerisedServiceControllerTypeBuilder
from useintest.services.controllers import DockerisedServiceController, ServiceType

ROOT_USERNAME = "root"
ROOT_PASSWORD = "gitlab123"

_repository = "gitlab/gitlab-ce"
_ports = [80, 433, 22]
_start_detector = lambda log_line: "==> /var/log/gitlab/redis/current <==" in log_line
_persistent_error_detector = lambda log_line: "o space left on device" in log_line
_environment_variables = {"GITLAB_ROOT_PASSWORD": ROOT_PASSWORD}


class GitLabBaseServiceController(DockerisedServiceController[ServiceType], metaclass=ABCMeta):
    """
    Base class for GitLab service controllers.
    """
    def start_service(self) -> DockerisedServiceWithUsers:
        service = super().start_service()
        service.root_user = User(ROOT_USERNAME, ROOT_PASSWORD)
        return service


common_setup = {
    "superclass": GitLabBaseServiceController,
    "repository": _repository,
    "start_detector": _start_detector,
    "persistent_error_detector": _persistent_error_detector,
    "ports": _ports,
    "additional_run_settings": {"environment": _environment_variables}
}

GitLab8_10_4_ce_0ServiceController = DockerisedServiceControllerTypeBuilder(
    name="GitLab8_10_4_ce_0ServiceController",
    tag="8.10.4-ce.0",
    **common_setup).build()

GitLab8_13_11_ce_0ServiceController = DockerisedServiceControllerTypeBuilder(
    name="GitLab8_13_11_ce_0ServiceController",
    tag="8.13.11-ce.0",
    **common_setup).build()

GitLab8_16_6_ce_0ServiceController = DockerisedServiceControllerTypeBuilder(
    name="GitLab8_16_6_ce_0ServiceController",
    tag="8.16.6-ce.0",
    **common_setup).build()

GitLab10_0_3_ce_0ServiceController = DockerisedServiceControllerTypeBuilder(
    name="GitLab10_0_3_ce_0ServiceController",
    tag="10.0.3-ce.0",
    **common_setup).build()


GitLabServiceController = GitLab10_0_3_ce_0ServiceController

gitlab_service_controllers = {GitLab8_10_4_ce_0ServiceController, GitLab8_13_11_ce_0ServiceController,
                              GitLab8_16_6_ce_0ServiceController, GitLab10_0_3_ce_0ServiceController}



import unittest
from abc import ABCMeta

from pymongo import MongoClient
from testhelpers import create_tests, TypeUsedInTest, get_classes_to_test

from useintest.modules.mongo import Mongo3DockerisedServiceController, MongoLatestDockerisedServiceController, \
    MongoServiceController, mongo_service_controllers
from useintest.services.models import DockerisedServiceWithUsers
from useintest.tests.services.common import TestDockerisedServiceControllerSubclass


class _TestMongoDockerisedServiceController(
        TestDockerisedServiceControllerSubclass[TypeUsedInTest, DockerisedServiceWithUsers], metaclass=ABCMeta):
    """
    Tests for Mongo service controllers.
    """
    def test_start(self):
        service = self._start_service()
        client = MongoClient(service.host, service.port)
        database = client["test-database"]
        posted = {"this": "value"}
        post_id = database.posts.insert_one(posted).inserted_id
        retrieved = database.posts.find_one({"_id": post_id})
        self.assertEqual(retrieved, posted)


# Setup tests
CLASSES_TO_TEST = {Mongo3DockerisedServiceController, MongoLatestDockerisedServiceController}
globals().update(create_tests(_TestMongoDockerisedServiceController, get_classes_to_test(mongo_service_controllers, MongoServiceController)))


# Fix for stupidity of test runners
del _TestMongoDockerisedServiceController, TestDockerisedServiceControllerSubclass, create_tests, get_classes_to_test


if __name__ == "__main__":
    unittest.main()



import argparse
import base64
import os
import sys
from copy import deepcopy
from typing import List, Iterable, Dict, Set, Callable, Any, Union

from dill import dill

from useintest.executables.common import CLI_ARGUMENTS

_PROJECT_DIRECTORY = os.path.join(os.path.dirname(os.path.abspath(__file__)), "../")
_ARGUMENTS_TO_MOUNT_SCRIPT = os.path.join(_PROJECT_DIRECTORY, "executables", "paths_to_mount.py")


class CommandsBuilder:
    """
    Builds commands to run an executable in Docker.
    """
    def __init__(self, executable: str=None, container: str=None, image: str=None, executable_arguments: List[str]=None,
                 get_path_arguments_to_mount: Callable[[List[Any]], Set[str]]=None,
                 ports: Dict[int, int]=None, mounts: Dict[str, Union[str, Set[str]]]=None,
                 variables: Iterable[str]=None, name: str=None, detached: bool=False, other_docker: str=""):
        self.executable = executable
        self.container = container
        self.image = image
        self.executable_arguments = executable_arguments
        self.get_path_arguments_to_mount = get_path_arguments_to_mount \
            if get_path_arguments_to_mount is not None else lambda arguments: set()
        self.ports = ports if ports is not None else dict()
        self.mounts = mounts if mounts is not None else dict()
        self.variables = variables if variables is not None else {}
        self.name = name
        self.detached = detached
        self.other_docker = other_docker

    def build(self) -> str:
        """
        Builds the commands.
        :return: build comamnds.
        """
        if self.container is None and self.image is None:
            raise ValueError("Must define either the Docker image or container to run commands in")
        if self.container is not None and self.image is not None:
            raise ValueError("Cannot build Docker command to work in for both an image and a running container")

        mounts = ""
        for local_volume, container_volumes in self.mounts.items():
            if not isinstance(container_volumes, set):
                container_volumes = {container_volumes}
            for container_volume in container_volumes:
                mounts += "-v %s:%s " % (local_volume, container_volume)

        ports = ""
        for local_port, container_port in self.ports.items():
            ports += "-p %d:%d" % (local_port, container_port)

        variables = ""
        for variable in self.variables:
            variables += "-e %s " % variable

        executable_arguments = " ".join(self.executable_arguments) if self.executable_arguments is not None else CLI_ARGUMENTS

        if self.get_path_arguments_to_mount is not None:
            serialised_arguments_parser = base64.b64encode(dill.dumps(self.get_path_arguments_to_mount)).decode("utf-8")

            calculate_additional_mounts = ("""
                $("%(python_interpreter)s" "%(python_arguments_script)s" "%(serialised_arguments_parser)s" %(cli_arguments)s)
            """ % {
                "python_interpreter": sys.executable,
                "python_arguments_script": _ARGUMENTS_TO_MOUNT_SCRIPT,
                "serialised_arguments_parser": serialised_arguments_parser,
                "cli_arguments": executable_arguments
            }).strip()
        else:
            calculate_additional_mounts = ""

        return """
            docker %(docker_noun)s -i \\
                %(name)s \\
                %(detached)s \\
                %(mounts)s %(calculate_additional_mounts)s \\
                %(ports)s \\
                %(variables)s \\
                %(other_docker)s \\
                %(image_or_container)s \\
                %(executable)s %(executable_arguments)s
        """ % {
            "calculate_additional_mounts": calculate_additional_mounts,
            "name": "--name %s" % self.name if self.name is not None else "",
            "detached": "-d" if self.detached else "",
            "mounts": mounts,
            "ports": ports,
            "variables": variables,
            "other_docker": self.other_docker,
            "docker_noun": "run" if self.image is not None else "exec",
            "image_or_container": self.image if self.image is not None else self.container,
            "executable": self.executable,
            "executable_arguments": executable_arguments
        }


class MountedArgumentParserBuilder:
    """
    TODO
    """
    ALL_POSITIONAL_ARGUMENTS = "*"

    def __init__(self, named_arguments: Set[str]=None, positional_arguments: Union[Set[int], str]=None):
        """
        TODO
        :param named_arguments:
        :param positional_arguments:
        """
        self.named_arguments = named_arguments if named_arguments is not None else set()
        self.positional_arguments = positional_arguments if positional_arguments is not None else set()

    def build(self) -> Callable[[List[Any]], Set[str]]:
        """
        TODO
        :return:
        """
        named_arguments = deepcopy(self.named_arguments)
        positional_arguments = deepcopy(self.positional_arguments)

        def get_mounts(cli_arguments: List[Any]) -> Set[str]:
            parser = argparse.ArgumentParser()
            for name in named_arguments:
                parser.add_argument(name, type=str)
            parser.add_argument("positional_arguments", type=str, nargs="*")
            arguments, _ = parser.parse_known_args(cli_arguments)

            mounts = set()  # type: Set[str]

            for name in named_arguments:
                value = getattr(arguments, name.lstrip("-"), None)
                if value is not None:
                    mounts.add(value)

            if positional_arguments != MountedArgumentParserBuilder.ALL_POSITIONAL_ARGUMENTS:
                for position in positional_arguments:
                    if position <= len(arguments.positional_arguments):
                        mounts.add(arguments.positional_arguments[position - 1])
            else:
                mounts = mounts.union(set(arguments.positional_arguments))

            return mounts

        return get_mounts
